/*
 * This software is copyright (c) 2004 by Shadowplay.net. It is released 
 * under the terms of the Artistic License.  Please refer to the file 
 * LICENSE for details.
 */
package net.shadowplay.wurml;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.shadowplay.http.ContentType;
import net.shadowplay.http.cookie.CookieJar;
import net.shadowplay.http.robot.RobotDb;
import net.shadowplay.wurml.filter.DiscoverIdFilter;
import net.shadowplay.wurml.filter.HrefFilter;
import net.shadowplay.wurml.filter.OptionOnpickFilter;
import net.shadowplay.wurml.input.ElementNameFilter;
import net.shadowplay.wurml.input.InputContainer;
import net.shadowplay.wurml.input.ItJoin;
import net.shadowplay.wurml.input.MalformedVariableException;
import net.shadowplay.wurml.input.SubstString;
import net.shadowplay.xml.DocumentBuilder;
import net.shadowplay.xml.DomParserFactory;
import net.shadowplay.xml.ParseErrorHandler;

import org.apache.xerces.dom.DocumentImpl;
import org.apache.xerces.parsers.DOMParser;
import org.w3c.dom.*;
import org.w3c.dom.traversal.NodeFilter;
import org.w3c.dom.traversal.NodeIterator;
import org.xml.sax.*;

import gnu.getopt.*;

/**
 * @author Michael Dykman for Shadowplay.net
 */
public class WurML implements Logger, ErrorCounter
{
	public static final int DEFAULT_DEPTH = 3;
	public static final String EOL = System.getProperty("line.separator");
	public static String [] ignoreMessages = new String []	{ };
	
	public static final String AGENT_NAME = "WurML"; 
	public static final String WURML_VERSION = "0.1.6";
	public static final String ROBOT_USER_AGENT = 
		AGENT_NAME  + "/" + WURML_VERSION
			+ " (+ http://wurml.shadowplay.net)"; 
	
	public static Map httpHeaders = 
		Collections.synchronizedMap(new HashMap());
	private static Map typeToMime = 
		Collections.synchronizedMap(new HashMap());
	private static Set ignoreURLs = 
		Collections.synchronizedSet(new HashSet());
	
	public static String [] acceptableImages = new String [] {};
	
	public static InputContainer inputContainer;

	private static RobotDb robotDb = new RobotDb(AGENT_NAME,ROBOT_USER_AGENT);
	
	public static String defaultEncoding = "UTF-8";

	public static final int OK = 0;
	public static final int REDIRECT = 1;
	public static final int FAILED = 2;
	
	public static boolean obeyRobots = true;
	public static boolean revisit = false;
	public static boolean imageCheck = true;
	public static boolean validate = true;
	public static boolean dryrun = false;
	public static boolean printLineNumbers = false;
	public static boolean showResponseHeaders = false;
	public static boolean keepLocal = false;
	public static boolean showPost = false;
	public static boolean strictEncoding = false;
	public static boolean doubleSpace = true;
	public static boolean brief = false;
	
	public static boolean reuseCookies = false;
	public static String defaultCookieFile = "wurml-cookies";
	public static String cookieFile = null;

	static
	{
		typeToMime.put(WapConstants.WML_TYPE,WapConstants.WML_MIME_TYPE);
		typeToMime.put(WapConstants.XHTML_TYPE,WapConstants.XHTML_MIME_TYPE);
		typeToMime.put(WapConstants.HTML_TYPE,WapConstants.HTML_MIME_TYPE);
	}
	static DocumentImpl configFactory = new DocumentImpl();
	private String warningURL = null;

	private DocumentImpl iteratorFactory = new DocumentImpl();
	private int depth;
	private 	String accepts = WapConstants.WML_MIME_TYPE;
	private 	String type = WapConstants.WML_TYPE;
	private 	String agent;
	private URL start = null;
	private boolean showCookies;
	private boolean dumpContent;
	private long elapsed = 0;
	
	protected int attempts = 0;
	protected int failures = 0;
	protected int warnings = 0;
	
	private Set visited;
	protected Map variables = new HashMap();
	protected DocumentBuilder parser;
	
	private CookieJar cookies;
	
	private static PrintWriter logWriter = null;
	private static String summaryLogName;
	
	public WurML(int depth, String type, String start, 
		boolean showCookies,boolean dumpContent)
		throws MalformedURLException, SAXException
	{
		HttpURLConnection.setFollowRedirects(false);
		cookies = new CookieJar();
		visited = new HashSet();
		this.depth = depth;
		this.type = type;
		this.accepts = (String) typeToMime.get(type);
		this.start = newURL(start);
		agent = AGENT_NAME;
		this.showCookies = showCookies;
		this.dumpContent = dumpContent;

		if(reuseCookies)
		{
			loadInitialCookies();
		}
		parser = DomParserFactory.createDocumentBuilder(
			type,validate, this, ignoreMessages);
	}
	
	protected static URL newURL(String s)
		throws MalformedURLException
	{
		URL url = new URL(s);
		return normalizePort(url);
	}
	
	protected static URL newURL(URL parent, String s)
		throws MalformedURLException
	{
		URL url = new URL(parent, s);
		return normalizePort(url);
	}
	
	protected static URL normalizePort(URL url)
		throws MalformedURLException
	{
		if(url.getPort() == url.getDefaultPort())
		{
			url = new URL(
			url.getProtocol(),
			url.getHost(),
			url.getFile()); 
		}
		return url;
	}
	
	protected String cookieFileName()
	{
		return cookieFile == null ? defaultCookieFile : cookieFile;
	}
	
	protected void saveCookies()
	{
		String file = cookieFileName();
		try
		{
			cookies.save(file);
		}
		catch(IOException e)
		{
			error("cannot save cookie file `" + file + "': " + e.getMessage());
		}
	}
	
	protected void loadInitialCookies()
	{
		
		String file = cookieFileName();
		try
		{
			cookies.load(file);
		}
		catch(IOException e)
		{
			error("cannot load cookie file `" + file + "': " + e.getMessage());
		}
	}
	
	protected void checkExternalLink(RequestProfile rp, int n)
	{
		Integer currentHash = new Integer(rp.hashCode());
		if( ++n > depth || ( !revisit && visited.contains(currentHash)))
		{
			return;
		}
		else
		{
			visited.add(currentHash);
		}
		CookieJar cj = new CookieJar();
		Response response = getResponse(rp, cj, -1);
		showContent(response);
	}
	
	public void incrementWarnings()
	{
		++warnings;
	}
	
	public void incrementErrors()
	{
		++failures;
	}
	
	protected void traverse(RequestProfile rp,int n)
	{
		int startcookies = cookies.hashCode();
		Integer currentHash = new Integer(rp.hashCode() ^ startcookies);
//System.out.println("traveral hash code = " + currentHash);		
		if( ++n > depth || ( !revisit && visited.contains(currentHash)))
		{
			return;
		}
		visited.add(currentHash);
		
		List complaints = new ArrayList();
		URL resource;

		Response response = getResponse(rp,cookies, n);

		if(response != null)
		{
			Document document = response.document;
			String encoding = response.encoding;
			resource = response.url;
			if(document != null)
			{
				if(imageCheck)
				{
					checkImages(resource,document,cookies);
				}
				
				RequestProfile[] urls = getURLs(resource,document);
				
				try
				{
					RequestProfile[] transforms = transformURLs(
						resource,document,urls, complaints, encoding);;
//	 I think all warnings have been reported by now
					showContent(response);
					for(int i = 0; i < transforms.length; ++i)
					{
						RequestProfile request = transforms[i];
						try
						{
							if(! checkIgnore(request.url))
							{
								URL url = request.createURL();
								if(dryrun)
								{
									logNL("found link: " + request.toExternalForm());
								}
								else
								{
									if(resource.getHost().equalsIgnoreCase(url.getHost()))
									{
										Integer starthc = new Integer(
											rp.toExternalForm().hashCode() 
											^ rp.paramHash()
											^ startcookies);
										visited.add(starthc);
										
										
										Integer hashCode = new Integer(
											url.toExternalForm().hashCode() 
											^ request.paramHash()
											^ cookies.hashCode());
										
										if(!visited.contains(starthc) && !visited.contains(hashCode))
										{
											if(! revisit)
											{
												visited.add(hashCode);
												visited.add(starthc);
											}
										}
										traverse(request,n);
									}
									else if(! keepLocal)
									{
										Integer hashCode = new Integer(
											url.toExternalForm().hashCode()
											^ request.paramHash());
										if(!visited.contains(hashCode))
										{
											if(! revisit)
											{
												visited.add(hashCode);
											}
										}
										checkExternalLink(request,n);
									}
								}
							}
						}
						catch(MalformedURLException e)
						{
							warning("invalid URL `" + request.url + "'");
						}
					}
				}
				catch(UnsupportedEncodingException e)
				{
					warning("Unsupported Encoding: " + e.getMessage());
				}
			}
		}
	}
	
	protected void checkImages(URL base, Document document, CookieJar cookys)
	{
		if(accepts.equals(WapConstants.WML_MIME_TYPE))
		{
			checkImagesOfType(base,document,cookys, acceptableImages);
		}
		else if(accepts.equals(WapConstants.XHTML_MIME_TYPE))
		{
			checkImagesOfType(base,document,cookys, null);
		}
	}

	protected void checkImagesOfType(
		URL base, 
		Document document, 
		CookieJar cookys, 
		String[] expectedMime)
	{
		NodeIterator it = iteratorFactory.createNodeIterator(document,NodeFilter.FILTER_ACCEPT,
			new ElementNameFilter("img"));
		Node n;
		Set checkedImages = new HashSet();
		
		while((n = it.nextNode()) != null)
		{
			NamedNodeMap attr = n.getAttributes();
			n = attr.getNamedItem("src");
			if(n == null)
			{
				warning("<IMG> tag found with no `src' attribute");
			}
			else
			{
				String img = n.getNodeValue();
				Node alt = attr.getNamedItem("alt");
				if(alt == null)
				{
					warning("<IMG src='" + img + "'> tag found without `alt' text");
				}
				
				if(!checkedImages.contains(img))
				{
					checkedImages.add(img);
					try
					{
						RequestProfile rp = new RequestProfile(RequestProfile.GET,
							base, img);
						URL imgURL = newURL(base,img);
						if(dryrun)
						{
							logNL("found image `" + imgURL.toExternalForm() + "'");
						}
						else
						{
							HttpURLConnection conn = submitURL(rp,cookys);
							imgURL = conn.getURL();
							int rc = conn.getResponseCode();
							if(rc !=  HttpURLConnection.HTTP_OK)
							{
								warning("loading image `" + imgURL.toExternalForm() 
									+ "' gives status code " + rc);
							}
							else
							{
								String mimeType = conn.getHeaderField("Content-Type");
								if(mimeType != null)
								{
									String mt = null;
									if(expectedMime != null)
									{
										for(int i = 0; i < expectedMime.length; ++i)
										{
											if(mimeType.equals(expectedMime[i]))
											{
												mt = expectedMime[i];
											}
										}
									}
									if(mt == null && expectedMime != null)
									{
										StringBuffer wb = new StringBuffer();
										wb.append("image `").append(imgURL.toExternalForm())
											.append("' reports Content-Type `").append(mimeType) 
											.append("' - should be ");
										if(expectedMime.length == 1)
										{
											wb.append("`").append(expectedMime[0]).append("'");
										}
										else
										{
											wb.append("one of ");
											for(int i = 0; i < expectedMime.length; ++i)
											{
												if(i > 0)
												{
													wb.append(", ");
												}
												wb.append("`").append(expectedMime[i]).append("'");
											}
										}
										
										warning(wb.toString());
									}
									else
									{
										XMLInputStream ic = 
											new XMLInputStream(conn.getInputStream(),this);
										int count = ic.consume();
										logNL("image `" + imgURL.toExternalForm() + 
											"' loaded successfully, content-type `" 
											+ mimeType + "', " 
											+ count + " bytes");
									}
								}
								else
								{
									warning("image `" + imgURL.toExternalForm() 
										+ "' gives no Content-Type");
								}
							}
						}
					}
					catch(MalformedURLException e)
					{
						warning("Malformed url `" + img + "' found in <IMG> tag");
					}
					catch(IOException e)
					{
						warning("IOException while fetching image `" + img + "': " + e.getMessage());
					}
				}
			}
		}
	}
	private static final void openSummaryLog(String name)
	{
		try
		{
			logWriter = new PrintWriter(new FileWriter(name));
		}
		catch(IOException e)
		{
			System.err.println("unable to open log `" + name + "' for writing: " 
				+ e.getMessage());
		}
	}
	public void traverse(RequestProfile rp)
	{
		traverse(rp,0);
			summaryLog(EOL + "urls attempted: " + getAttempts() 
				+ ", failures: " + getFailures()
				+ ", warnings:" + getWarnings() + EOL);
	}
	public  void summaryLog(String msg)
	{
		if(logWriter != null)
		{
			logWriter.write(msg);
		}
	}

	public int getAttempts()
	{
		return attempts;
	}

	public int getFailures()
	{
		return failures;
	}

	public int getWarnings()
	{
		return warnings;
	}

	protected RequestProfile[] transformURLs(URL parent, Document document,
		RequestProfile[] rp,List complaints, String encoding)
		throws UnsupportedEncodingException
	{
		List list = new ArrayList();
		for(int i = 0; i < rp.length; ++i)
		{
			RequestProfile [] rps = transformURLs(parent,document,rp[i],complaints, encoding);
			for(int j = 0; j < rps.length; ++j)
			{
				list.add(rps[j]);
			}
		}
		return (RequestProfile[]) list.toArray(new RequestProfile[list.size()]);
	}

	protected RequestProfile[] transformURLs(URL parent, Document document,
		RequestProfile rp,List complaints, String encoding)
		throws UnsupportedEncodingException
	{
		int substType, valueType;
		if(type.equals(WapConstants.WML_TYPE))
		{
			substType = SubstString.MODE_MIXED;
			valueType = SubstString.MODE_MIXED;
		}
		else// if(type.equals(WapConstants.XHTML_TYPE))
		{
			substType = SubstString.MODE_RAW;
			valueType = SubstString.MODE_VAR;
		}
		
		List transforms = new ArrayList();
		RequestProfile request;
		try
		{
			SubstString subst = 
				new SubstString(this,parent,document,variables,rp.url, complaints,type,substType,encoding);
			Iterator it = subst.iterator();
			if(rp.params.size() == 0)
			{
				while(it.hasNext())
				{
					String url = it.next().toString();
					request = new RequestProfile(rp.method,
						parent, url);
					transforms.add(request);
				}
			}
			else
			{
				ItJoin.Pair [][] params = new ItJoin.Pair[rp.params.size()][];
				int [] indices = new int [rp.params.size()];
				List cmb;
				
				Iterator pit = rp.params.iterator();
				int index = 0;
				while(pit.hasNext())
				{
					Param param = (Param) pit.next();
					cmb = new ArrayList();
					SubstString nst =	new SubstString(this,parent,param.searchNode,
						variables,param.name, complaints,type,substType, encoding);
					SubstString vst;
					if(valueType == SubstString.MODE_VAR)
					{
						vst = new SubstString(this,parent,param.searchNode,
							variables,param.name, complaints,type,valueType, encoding);
					}
					else  // valueType == SubstString.MIXED
					{
						vst = new SubstString(this,parent,param.searchNode,
							variables,param.value, complaints,type,valueType,encoding);
					}
					ItJoin j = new ItJoin(nst,vst);

					while(j.hasNext())
					{
						cmb.add(j.next());
					}
					params[index] = (ItJoin.Pair[]) cmb.toArray(new ItJoin.Pair [cmb.size()]);
					if(params[index].length == 0)
					{
						throw new MalformedVariableException("cannot find a substitution for `" 
							+ param.name + '=' + param.value + "'");
					}
					++index;
				}
				while(it.hasNext())
				{
					String url = it.next().toString();
					for(int i = 0; i < indices.length; ++i)
					{
						indices[i] = 0;
					}
					boolean done = false;
					
					while(!done)
					{
						request = new RequestProfile(rp.method,
							parent, url);
						for(int i =  0; i < indices.length ; ++i)
						{
							ItJoin.Pair p = params[i][indices[i]];
// add literal params							
							request.addParam(p.first.toString(),p.second.toString(),null);
						}
						indices[indices.length-1]++;
						transforms.add(request);
						for(int i = indices.length - 1; i >= 0 ; --i)
						{
							if(indices[i] == params[i].length)
							{
								indices[i] = 0;
								if(i > 0)
								{
									indices[i-1]++;
								}
								else
								{
									done = true;
								}
							}
							else
							{
								break;
							}
						}
					}
				}
			}
				
		}
		catch(MalformedVariableException e)
		{
			warning(e.getMessage());
		}
		return (RequestProfile[]) transforms.toArray(
			new RequestProfile[transforms.size()]);
	}
	
	protected boolean checkIgnore(String url)
	{
		Iterator it = ignoreURLs.iterator();
		boolean result = false;
		while(it.hasNext())
		{
			if(url.startsWith(it.next().toString()))
			{
				result = true;
				break;
			}
		}
		return result;
	}

	protected boolean robotsNoFollow(URL base, Document document)
	{
		boolean result = false;

		NodeIterator it;
		Node node;
		
		if(obeyRobots)
		{
			it = iteratorFactory.createNodeIterator(
				document,NodeFilter.FILTER_ACCEPT,new ElementNameFilter("META"));
			
			while((node = it.nextNode()) != null)
			{
				Node nn;
				NamedNodeMap attr = node.getAttributes();
				if((nn = attr.getNamedItem("name")) != null)
				{
					if(nn.getNodeValue().equalsIgnoreCase("robots"))
					{
						nn = attr.getNamedItem("content");
						if(nn != null)
						{
							String content = nn.getNodeValue();
							if((content.indexOf("nofollow") >= 0)
								|| (content.indexOf("NOFOLLOW") >= 0))
							{
								logNL("not following any urls loaded from `"
									+ base.toExternalForm() + "' as per <META name=\"robots\">");
								result = true;
							}
						}
					}
				}
			}
		}
		return result;
	}

	protected RequestProfile[] getURLs(URL base, Document document)
	{
		RequestProfile[] urls = new RequestProfile[] {};
		
		if(! robotsNoFollow(base,document))
		{
			if(type.equals(WapConstants.WML_TYPE))
			{
				urls = getURLsFromWML(base,document);
			}
			else if(type.equals(WapConstants.XHTML_TYPE))
			{
				urls = getURLsFromXHTML(base,document);
			}
		}
		
		return urls;
	}
	
	protected RequestProfile[] getURLsFromXHTML(URL base, Document document)
	{
		List rps = new ArrayList();
		NodeIterator it = iteratorFactory.createNodeIterator(
			document,NodeFilter.FILTER_ACCEPT,new ElementNameFilter("HREF"));
		Node node;
		RequestProfile rp;
		while((node = it.nextNode()) != null)
		{
			NamedNodeMap attr = node.getAttributes();
			Node href = attr.getNamedItem("href");
			if(href != null)
			{
				rp = new RequestProfile(RequestProfile.GET,base,href.getNodeValue());
				rps.add(rp);
			}
		}
		it = iteratorFactory.createNodeIterator(
			document,NodeFilter.FILTER_ACCEPT,new ElementNameFilter("FORM"));
		while((node = it.nextNode()) != null)
		{
		}
		
		return (RequestProfile[]) rps.toArray(new RequestProfile[rps.size()]);
	}

	protected RequestProfile[] getURLsFromWML(URL base, Document document)
	{
		NodeIterator it;
		Node node;

		List urls = new ArrayList();
		RequestProfile rp;
		
		it = iteratorFactory.createNodeIterator(
				document,NodeFilter.FILTER_ACCEPT,new OptionOnpickFilter());
		while((node = it.nextNode()) != null)
		{
			String href = node.getAttributes().getNamedItem("onpick").getNodeValue();
			rp = new RequestProfile(RequestProfile.GET,base,href);
			addURL(document,node,urls,rp);
		}

		it = iteratorFactory.createNodeIterator(document,NodeFilter.FILTER_ACCEPT,
			new ElementNameFilter("CARD"));
		while((node = it.nextNode()) != null)
		{
			NamedNodeMap attr = node.getAttributes();
			Node n;
			if((n =	attr.getNamedItem("ontimer")) != null)
			{
				rp = new RequestProfile(RequestProfile.GET,base,n.getNodeValue());
				addURL(document,node,urls,rp);
			}
			if((n =	attr.getNamedItem("onenterforward")) != null)
			{
				rp = new RequestProfile(RequestProfile.GET,base,n.getNodeValue());
				addURL(document,node,urls,rp);
			}
			if((n =	attr.getNamedItem("onenterbackward")) != null)
			{
				rp = new RequestProfile(RequestProfile.GET,base,n.getNodeValue());
				addURL(document,node,urls,rp);
			}
		}
		
		it = iteratorFactory.createNodeIterator(document,NodeFilter.FILTER_ACCEPT, new HrefFilter());
		while((node = it.nextNode()) != null)
		{
			NamedNodeMap attr = node.getAttributes();
			String href = attr.getNamedItem("href").getNodeValue();
			Node method = attr.getNamedItem("method");
			if(node.getNodeName().equalsIgnoreCase("GO"))
			{
				if(method == null)
				{
					rp = new RequestProfile(RequestProfile.GET,base,href);
				}
				else
				{
					rp = new RequestProfile(method.getNodeValue().toUpperCase(),
						base,href);
				}
				NodeIterator pfIt = iteratorFactory.createNodeIterator(node,
					NodeFilter.FILTER_ACCEPT, new ElementNameFilter("postfield"));
				Node pf;
				while((pf = pfIt.nextNode()) != null)
				{
					attr = pf.getAttributes();
					rp.addParam(attr.getNamedItem("name").getNodeValue(),
						attr.getNamedItem("value").getNodeValue(),document);
				}
			}
			else
			{
				rp = new RequestProfile(RequestProfile.GET,base, href);
			}
			
			addURL(document,node,urls,rp);
		}
		
		return (RequestProfile[]) urls.toArray(new RequestProfile[urls.size()]);
	}
	
	public void addURL(Document document, Node node,List list, RequestProfile rp)
	{
		int offset = rp.url.indexOf('#');
		if(offset == 0)
		{
			checkLocalAnchor(document,node,rp.url.substring(1));
		}
		else
		{
			if(offset != -1)
			{
				rp.url = rp.url.substring(0,offset);
			}
			
			list.add(rp);
		}
	}
	
	public void setAgent(String agent)
	{
		this.agent = agent;
	}
	
	public static int evaluateResponseCode(int code)
	{
		int result = FAILED;
		if(code >= 200 && code < 300)
		{
			result = OK;
		}
		else if(code == 302 || code == 301)
		{
			result = REDIRECT;
		}
		
		return result;
	}
	
	protected static boolean mimeFit(String type, String receive)
	{
		boolean result = false;
		if(type.equals(WapConstants.WML_TYPE))
		{
			result = receive.startsWith(WapConstants.WML_MIME_TYPE);
		}
		else if(type.equals(WapConstants.XHTML_TYPE))
		{
			if((receive.indexOf("text/html") != -1) ||
				(receive.indexOf("application/xhtml+xml") != -1) ||
				(receive.indexOf("application/vnd.wap.xhtml+xml") != -1))
			{
				result = true;
			}
		}
		return result;
	}
	private void setRequestHeaders(HttpURLConnection conn, CookieJar cj)
	{
		conn.setRequestProperty("Accept",accepts);
		if(cj != null)
		{
			String cc = cj.cookieString();
			if(cc != null)
			{
				if(showCookies)
				{
					logNL("Cookie: " + cc);
				}
				conn.setRequestProperty("Cookie",cc);
			}
		}
		Iterator it = httpHeaders.keySet().iterator();
		while(it.hasNext())
		{
			Object key = it.next();
			conn.setRequestProperty(
				key.toString(),httpHeaders.get(key).toString());
		}
		conn.setRequestProperty("User-Agent",agent);
	}
	
	protected Response getResponse(
		RequestProfile rp,
		CookieJar cookys,
		int dep)
	{
		Response response = null;
		boolean allowed = true;
		URL current;
		if(doubleSpace)
		{
			logNL("");
		}
		
		try
		{
			current = rp.createURL();
			warningURL = rp.toExternalForm();

			if(obeyRobots)
			{
				try
				{
					allowed = robotDb.allow(current);
				}
				catch(IOException e)
				{
					warning("IOException loading robots.txt for " + rp.url);
				}
			}
			else
			{
				allowed = true;
			}
		}
		catch(MalformedURLException e)
		{
			warning("MalformedURLException creating link for " + rp.url);
			return null;
		}
		
		if(allowed)
		{
			++attempts;
			response =  fetchContent(rp,current,cookys,dep);
			if(response == null || response.document == null)
			{
				incrementErrors();
			}
		}
		else
		{
			logNL("skipping `" + rp.url 
				+ "' as per robot text");
		}
		return response;
	}
	protected Response fetchContent(
		RequestProfile rp,
		URL current,
		CookieJar cookys,
		int dep)
	{
		Document document = null;
		Response response  = null;
		
		XMLInputStream cis = null;
		String encoding = null;

		try
		{
			showPostParams(rp);
			HttpURLConnection http = submitURL(rp,cookys);
			if(http == null)
			{
				return response;
			}
			current = http.getURL();
			log(current.toExternalForm() + " ");
			response = new Response(current, null, encoding);

			int status = http.getResponseCode();
			int r = evaluateResponseCode(status);
			
			if(r == FAILED)
			{
				error("status " + status + " - failed to load");
			}
			else
			{
				ParseErrorHandler errorHandler = parser.getParseErrorHandler();
				errorHandler.setDoc(current.toExternalForm());
				String mimeType = http.getHeaderField("Content-Type");
				ContentType ct = null;
				if(mimeType != null)
				{
					ct = new ContentType(mimeType);
					mimeType = ct.getMime();
				}
				cis = new XMLInputStream(http.getInputStream(),this);
				cis.consume();
				elapsed = System.currentTimeMillis() - elapsed;
				byte[] bytes = cis.getBuffer();
				int leading = getLeadingSpace(bytes);
				
				encoding = solveEncoding(ct, cis,leading);
				response.encoding = encoding;
				response.headers = http.getHeaderFields();
				response.text = cis.getText(encoding);

				InputSource inputSource = new InputSource(
					new ByteArrayInputStream(bytes,leading, bytes.length - leading));
				inputSource.setEncoding(encoding);
				inputSource.setSystemId(current.toExternalForm());
				
				try
				{
					parser.parse(inputSource);
					
					document = parser.getDocument();
					response.document = document;
					DocumentType doctype = document.getDoctype();
					String typeName = doctype.getName();

					if(errorHandler.errorFlag())
					{
						if(brief)
						{
							logNL("- Syntax Warning: " + errorHandler.toString());
						}
						else
						{
							logNL(errorHandler.toString() + "- elapsed time " 
								+ elapsed + " ms");
						}
						errorReport(rp.url,rp.parent,errorHandler,cis.getBuffer(),
							encoding,elapsed);
						incrementErrors();
					}
					else
					{
						if(brief)
						{
							logNL( "- OK");
						}
						else
						{
							logNL("- " + status + " - " + typeName 
								+ " - " + cis.getCount() 
								+ " bytes - validates OK - elapsed time " 
								+ elapsed + " ms "  + "(depth:" + dep + ")");
						}
					}
					if(leading > 0)
					{
						warning("leading spaces at the start of the document");
					}
					
					if(! typeToMime.containsKey(typeName))
					{
						document = null;
					}
					checkDocType(typeName);
					if(mimeType == null)
					{
						warning("no content type specified in response.");
					}
					else if(! mimeFit(type, mimeType))
					{
						warning("unexpected content type specified in response. `"
							+ mimeType + "'");
					}
				}
				catch(SAXException e)
				{
					if(brief)
					{
						logNL("- Syntax Error: " + e.getMessage());
					}
					else
					{
						logNL(start.toExternalForm() + " Status " + status 
							+ " (" + cis.getCount() + " bytes), validation error "  
							+ e.getMessage()
							+ " - elapsed time " + elapsed + " ms");
					}
					errorReport(rp.url,rp.parent,errorHandler,cis.getBuffer(),
						encoding,elapsed);
					incrementErrors();
				}
			}
		}
		catch(UnsupportedEncodingException e)
		{
			StringBuffer err = new StringBuffer();
			err.append("- Unsupported Character Encoding: ")
				.append(e.getMessage());
			if(!brief)
			{
				err.append("- Unsupported Character Encoding: ")
					.append(e.getMessage());
				elapsed = System.currentTimeMillis() - elapsed;
			}
			logNL(err.toString());
			errorReport(start.toExternalForm(),rp.parent,e,elapsed);
			incrementErrors();
		}
		catch(IOException e)
		{
			StringBuffer err = new StringBuffer(rp.url);
			err.append("- IOException while loading document - ")
				.append(e.getMessage());
			if(!brief)
			{
				elapsed = System.currentTimeMillis() - elapsed;
			}
			logNL(err.toString());
			errorReport(start.toExternalForm(),rp.parent,e,elapsed);
			incrementErrors();
		}

		return response;
	}
	protected void showContent(Response response)
	{
		showResponseHeaders(response.headers);
		if(dumpContent)
		{
			logNL(response.text);
		}
	}
	protected String solveEncoding(ContentType ct,XMLInputStream xis, int offset)
		throws UnsupportedEncodingException
	{
		String encoding;
		String mime = ct.getMime();
		encoding = ct.getEncoding();
		if(encoding == null)
		{
			if(strictEncoding && mime.startsWith("text/"))
			{
				encoding = WapConstants.DEFAULT_TEXT_ENCODING;
			}
			else
			{
				encoding = xis.guessEncoding(offset);
				if(encoding == null)
				{
					encoding = strictEncoding ?
						WapConstants.DEFAULT_APPLICATION_ENCODING
						: defaultEncoding;
				}
			}
		}
		return encoding;
	}
	
	protected void showPostParams(RequestProfile rp)
	{
		if(showPost && rp.method.equals(RequestProfile.POST) && rp.params.size() > 0)
		{
			Param p;
			Iterator it = rp.params.iterator();
			while(it.hasNext())
			{
				p = (Param) it.next();
				logNL("[post param] " + p.name + "=" + p.value);
			}
		}
	}
	protected void checkDocType(String doctype)
	{
		if(doctype == null)
		{
			warning("no DTD present, doctype is null");
		}
		else if(type.equals(WapConstants.WML_TYPE))
		{
			if(! doctype.equals(type))
			{
				warning("unexpected document type `" + doctype + "'");
			}
		}
		else if(type.equals(WapConstants.XHTML_TYPE))
		{
			if(! doctype.equals(WapConstants.HTML_TYPE))
			{
				warning("unexpected document type `" + doctype + "'");
			}
		}
	}
	
	protected int getLeadingSpace(byte[] bytes)
	{
		int i = 0;
		for(; i < bytes.length && Character.isWhitespace((char)bytes[i]); ++i)
			{ }
		return i;
	}
	
	protected HttpURLConnection submitURL(
		RequestProfile rp,
		CookieJar cookys)
		throws IOException
	{
		Set redirected = new HashSet();
		HttpURLConnection http = null;
		int status;
		int r = FAILED;
		String rdu = null;
		
		try
		{
			do
			{
				Integer linkHash = new Integer(rp.hashCode() ^ cookys.hashCode());
				if(redirected.contains(linkHash))
				{
					warning("redirect loop detected:");
					Iterator it = redirected.iterator();
					while(it.hasNext())
					{
						warning("\t" + ((URL)it.next()).toExternalForm(),false);
					}
					return null;
				}
				else
				{
					URL url;
					redirected.add(linkHash);
					StringBuffer requestContent= new StringBuffer();
					if(rp.params.size() > 0)
					{
						Param p;
						Iterator it = rp.params.iterator();
						while(it.hasNext())
						{
							p = (Param) it.next();
							if(p.value != null)
							{
								if(requestContent.length() > 0)
								{
									requestContent.append('&');
								}
								requestContent.append(URLEncoder.encode(p.name,"UTF-8"))
									.append('=').append(URLEncoder.encode(p.value,"UTF-8"));
							}
						}
					}
					if(rp.method.equals(RequestProfile.GET) && rp.params.size() > 0)
					{
						StringBuffer uu = new StringBuffer(rp.url);
						if(uu.indexOf("?")  == -1)
						{
							uu.append('?');
						}
						else
						{
							uu.append('&');
						}
						uu.append(requestContent);
						url = newURL(rp.parent,uu.toString());
					}
					else
					{
						url = rp.createURL();
					}
					
					URLConnection uc = url.openConnection();
					if(! (uc instanceof HttpURLConnection))
					{
						warning("can not follow `" + url + "', only HTTP supported");
						return null;
					}
					http = (HttpURLConnection) uc;
					http.setUseCaches(false);
					http.setRequestMethod(rp.method);
					setRequestHeaders(http,cookys);
					
					if(rp.method.equals(RequestProfile.POST) && rp.params.size() > 0)
					{
						http.setRequestProperty("Content-Encoding",
							WapConstants.POST_ENCODING);
						http.setDoOutput(true);
						http.setRequestProperty("Content-Length",
							Integer.toString(requestContent.length()));
						OutputStream rout = http.getOutputStream();
						rout.write(requestContent.toString().getBytes());
						rout.flush();
					}

					
					elapsed = System.currentTimeMillis();
					status = http.getResponseCode();
					Map headers = http.getHeaderFields();
					cookys.addCookies(headers,showCookies);
					r = evaluateResponseCode(status);
	
					
					if(r == REDIRECT)
					{
						showPostParams(rp);
						showResponseHeaders(headers);
						
						rdu = uc.getHeaderField("Location");
						if(rdu == null)
						{
							throw new IOException("Redirect request (status " + status +") sent but no Location headers found");
						}
						elapsed = System.currentTimeMillis()  - elapsed;
						logNL("redirecting `" + rp.url + "' to `" + rdu + "' - elapsed time " + elapsed + " ms");
						rp = new RequestProfile(RequestProfile.GET,url,rdu);
					}
				}
			} while(r == REDIRECT);
		}
		catch(MalformedURLException e)
		{
			http = null;
			warning("malformed url in redirect: `" + rdu + "'");
		}
		
		return http;
	}
	protected void showResponseHeaders(Map headers)
	{
		if(showResponseHeaders)
		{
			StringBuffer buff;
			Iterator it = headers.keySet().iterator();
			logNL("");
			while(it.hasNext())
			{
				Object h = it.next();
				Object v = headers.get(h);
				if(v != null)
				{
					List vals = (List) v;
					Iterator vit = vals.iterator();
					while(vit.hasNext())
					{
						buff = new StringBuffer();
						if(h != null)
						{
							buff.append((String)h).append(": ");
						}
						String value = ((String)vit.next());
						buff.append(value);
						logNL(buff.toString());
					}
				}
			}
			logNL("");
		}

	}
	protected void checkLocalAnchor(Document document, Node nn, String id)
	{
		String parent = nn.getNodeName();
		if(id.trim().length() == 0)
		{
			warning("empty local reference in <" + parent + ">");
		}
		NodeIterator it = iteratorFactory.createNodeIterator(document,
			NodeFilter.FILTER_ACCEPT, new DiscoverIdFilter(id));
		Node node = it.nextNode();
		if(node == null)
		{
			warning("`#" + id + "' used as target in <" + parent + "> but no tag found with that id"); 
		}
		else
		{
			if(! node.getNodeName().equalsIgnoreCase("CARD"))
			{
				warning("target of local reference `#" + id + 
					"' in <" + parent + "> is not a <CARD>"); 
			}
			if(it.nextNode() != null)
			{
				warning("more than one tag with id `" + id + "'"); 
			}
		}
	}
	private void errorReport(String title,URL parent,String content)
	{
		summaryLog(title + EOL);
		if(parent != null)
		{
			summaryLog("\tin document from url" + parent.toExternalForm() + EOL);
				
		}
		summaryLog(content);
		summaryLog(EOL + 
			"========================= END ERROR REPORT =========================" + EOL);
	}
	private void errorReport(String url,URL parent, IOException e, long elapsed)
	{
		StringBuffer buffer = new StringBuffer();
		StringWriter sw = new StringWriter();
		e.printStackTrace(new PrintWriter(sw));
		buffer.append("IO error while fetching url ").append(url);
		if(parent != null)
		{
			buffer.append(" from ").append(parent.toExternalForm());
		}
		buffer.append(" - elapsed time ")
			.append(elapsed).append(" ms")
			.append(EOL).append(EOL)
			.append(e.getMessage()).append(EOL)
			.append(sw.getBuffer()).append(EOL);
		errorReport("Error: io-error in " + url,parent,buffer.toString());
	}
	
	private void errorReport(String doc,URL parent,ParseErrorHandler eh, byte[] content,String encoding, long elapsed)
	{
		StringBuffer buffer = new StringBuffer(); 
		buffer.append(eh.toString())
			.append("elapsed time ")
			.append(elapsed).append(" ms, ");

		String cs;
		if(encoding != null)
		{
			try
			{
				cs = new String(content,encoding);
			}
			catch(UnsupportedEncodingException e)
			{
				buffer.append("Content encoding `").append(encoding).
					append("' not supported on this system. Using system default instead");
				cs = new String(content);
			}
		}
		else
		{
			cs = new String(content);
		}
		buffer.append("======================== document received =========================")
			.append(EOL);
		BufferedReader reader = new BufferedReader(new StringReader(cs));
		String s;
		int lineno = 0;
		if(printLineNumbers)
		{
			try
			{
				while((s = reader.readLine()) != null)
				{
					buffer.append(++lineno).append(": ").append(s).append(EOL);
				}
			}
			catch(IOException e)
			{
				buffer.append("unable to write content");
			}
		}
		else
		{
			buffer.append(cs).append(EOL);
		}
		
		buffer.append(EOL);
		
		errorReport("Error: syntax-error in " + doc,parent,buffer.toString());
	}
	
	public static void log(String msg)
	{
System.out.print(msg);		
	}
	public static void logNL(String msg)
	{
System.out.println(msg);		
	}
	public void warning(String msg)
	{
		warning(msg,true);
	}
	public void error(String msg)
	{
		logNL("Error: " + msg);	
		summaryLog("Error: " + msg + EOL);
	}
	public void warning(String msg, boolean inc)
	{
		if(warningURL != null)
		{
			summaryLog("in " + warningURL  + ": ");
			warningURL = null;
		}
		
		if(inc)
		{
			++warnings;
			logNL("Warning: " + msg);	
			summaryLog("Warning: " + msg + EOL);
		}
		else
		{
			logNL(msg);	
			summaryLog(msg + EOL);
		}
	}
	private static String makeAttributeIgnoreMessage(String tag, String attribute)
	{
		StringBuffer buffer = new StringBuffer();
		buffer.append("Attribute \"").append(attribute)
			.append("\" must be declared for element type \"")
			.append(tag).append('"');
		return buffer.toString();
	}
	
	public static void setContentEncoding(Document document)
	{
		NodeIterator it = configFactory.createNodeIterator(
			document, NodeFilter.FILTER_ACCEPT, new ElementNameFilter("encoding"));
		Node n = it.nextNode();
		if(n != null)
		{
			defaultEncoding = n.getFirstChild().getNodeValue();
		}
	}

	public static void setIgnores(Document document)
	{
		NodeIterator it = configFactory.createNodeIterator(
			document, NodeFilter.FILTER_ACCEPT, new ElementNameFilter("ignore-url"));
		Node n;
		while((n = it.nextNode()) != null)
		{
			ignoreURLs.add(n.getFirstChild().getNodeValue());
		}
	}

	public static void setSyntaxFilter(Document document)
	{
		List list = new ArrayList();
		NodeIterator it = configFactory.createNodeIterator(
			document, NodeFilter.FILTER_ACCEPT, new ElementNameFilter("xml-syntax"));
		Node n;
		while((n = it.nextNode()) != null)
		{
			NodeIterator eit = configFactory.createNodeIterator(
				n, NodeFilter.FILTER_ACCEPT, new ElementNameFilter("element"));
			while((n = eit.nextNode()) != null)
			{
				String ename, aname;
				
				NamedNodeMap attr = n.getAttributes();
				Node nn = attr.getNamedItem("name");
				ename = nn.getNodeValue();
				NodeIterator ait = configFactory.createNodeIterator(
					n, NodeFilter.FILTER_ACCEPT, new ElementNameFilter("ignore-attribute"));
				while((n = ait.nextNode()) != null)
				{
					attr = n.getAttributes();
					nn = attr.getNamedItem("name");
					aname = nn.getNodeValue();
					list.add(makeAttributeIgnoreMessage(ename,aname));
				}
			}
		}
		ignoreMessages = (String[]) list.toArray(new String[list.size()]);
	}
	
	public static void setAcceptImages(Document document)
	{
		NodeIterator it = configFactory.createNodeIterator(
			document, NodeFilter.FILTER_ACCEPT, new ElementNameFilter("header"));
		Node n;
		String value;
		Set image = new HashSet();
		
		image.add(WapConstants.WML_IMAGE_TYPE);
		
		while((n = it.nextNode()) != null)
		{
			NodeList nl = n.getChildNodes();
			for(int i = 0; i < nl.getLength(); ++i)
			{
				n = nl.item(i);
				if(n.getNodeType() != Node.COMMENT_NODE)
				{
					if(n.getNodeName().equalsIgnoreCase("allow"))
					{
						value = n.getFirstChild().getNodeValue();
						image.add(value);
					}
				}
			}
		}
		acceptableImages = (String[]) image.toArray(new String[image.size()]); 
	}
	public static void setHeaders(Document document)
	{
		NodeIterator it = configFactory.createNodeIterator(
			document, NodeFilter.FILTER_ACCEPT, new ElementNameFilter("header"));
		Node n;
		String name = null, value;
		while((n = it.nextNode()) != null)
		{
			NodeList nl = n.getChildNodes();
			for(int i = 0; i < nl.getLength(); ++i)
			{
				n = nl.item(i);
				if(n.getNodeType() != Node.COMMENT_NODE)
				{
					if(n.getNodeName().equalsIgnoreCase("name"))
					{
						name = n.getFirstChild().getNodeValue();
					}
					else if(n.getNodeName().equalsIgnoreCase("value"))
					{
						value = n.getFirstChild().getNodeValue();
						httpHeaders.put(name,value);
						break;
					}
				}
			}
		}
	}
	
	public static void setProxy(String proxy)
		throws MalformedURLException
	{
		int offset = proxy.indexOf(':');
		if(offset == -1)
		{
			System.setProperty("http.proxyHost",proxy);
		}
		else
		{
			String host = proxy.substring(0,offset);
			System.setProperty("http.proxyHost",host);
			if(proxy.length() > offset + 1)
			{
				String port = proxy.substring(offset+1);
				System.setProperty("http.proxyPort",port);
			}
		}
	}
	
	public static void main(String[] args)
	{
		int depth = DEFAULT_DEPTH;
		String type = "wml";
		String agent = null;
		boolean showCookies = false;
		boolean dumpContent = false;
		boolean validate_config = true;
		String conf = null;
		String proxy = null;
		 
		boolean opterror =false;
		LongOpt[] longopts = new LongOpt[] {
			new LongOpt("type", LongOpt.REQUIRED_ARGUMENT, null, 't'),
			new LongOpt("depth", LongOpt.REQUIRED_ARGUMENT, null, 'd'), 
			new LongOpt("agent", LongOpt.REQUIRED_ARGUMENT, null, 'g'),
			new LongOpt("show-cookies", LongOpt.NO_ARGUMENT, null, 's'), 
			new LongOpt("dump", LongOpt.NO_ARGUMENT, null, 'm'),
			new LongOpt("config", LongOpt.REQUIRED_ARGUMENT, null, 'c'), 
			new LongOpt("usage", LongOpt.NO_ARGUMENT, null, 'u'),
			new LongOpt("ignore-robots", LongOpt.NO_ARGUMENT, null, 'i'), 
			new LongOpt("revisit", LongOpt.NO_ARGUMENT, null, 'r'),
			new LongOpt("version", LongOpt.NO_ARGUMENT, null, 'v'), 
			new LongOpt("proxy", LongOpt.REQUIRED_ARGUMENT, null, 'p'), 
			new LongOpt("no-validation", LongOpt.NO_ARGUMENT, null, 'n'),
			new LongOpt("dry-run", LongOpt.NO_ARGUMENT, null, 'y'),
			new LongOpt("trust-config", LongOpt.NO_ARGUMENT, null, 'x'),
			new LongOpt("lines", LongOpt.NO_ARGUMENT, null, 'l'),
			new LongOpt("show-headers", LongOpt.NO_ARGUMENT, null, 'h'),
			new LongOpt("keep-local", LongOpt.NO_ARGUMENT, null, 'k'),
			new LongOpt("show-post", LongOpt.NO_ARGUMENT, null, 'P'),
			new LongOpt("strict-encoding", LongOpt.NO_ARGUMENT, null, 'E'),
			new LongOpt("single-space", LongOpt.NO_ARGUMENT, null, 'S'),
			new LongOpt("saved-cookies", LongOpt.OPTIONAL_ARGUMENT, null, 'C'), 
			new LongOpt("brief", LongOpt.NO_ARGUMENT, null, 'b'), 
		};
		
		Getopt g = new Getopt("WapWorm", args, "PESC::bkhlxnriumst:d:a:c:p:",longopts);
		g.setOpterr(false);
		int c;
		while ((c = g.getopt()) != -1)
		{
			switch(c)
			{
		 		case 't' :
		 			type = g.getOptarg();
		 		break;

			 	case 'd' :
		 			depth = Integer.parseInt(g.getOptarg());
			 	break;
			 	
			 	case 'a' :
			 		agent = g.getOptarg();
			 	break;
			 	
			 	case 's' : 
			 		showCookies = true;
			 	break;
			 	
			 	case 'm' : 
			 		dumpContent = true;
			 	break;
			 	
			 	case 'c' : 
			 		conf = g.getOptarg();
			 	break;

			 	case 'u' : 
			 		showUsage();
			 		System.exit(0);
			 	break;

			 	case 'i' : 
			 			obeyRobots = false;
			 	break;

			 	case 'r' : 
			 			revisit = true;
			 	break;
			 	
			 	case 'v' : 
System.out.println(AGENT_NAME + " version " + WURML_VERSION
	+ " (c) 2004-2005 Shadowplay.net ( http://www.shadowplay.net )");
System.out.print("Released under the Artistic License.");
System.out.println("  author: Michael Dykman.");
System.exit(0);
			 	break;
			 	
			 	case 'p' :
			 		proxy = g.getOptarg();
			 	break;
			 	
			 	case 'n' :
			 		validate = false;
			 	break;
			 	
			 	case 'y' :
			 		dryrun = true;
			 	break;
			 	
			 	case 'x' :
			 		validate_config = false;
			 	break;
			 	
			 	case 'l' :
			 		printLineNumbers = true;
			 	break;
			 	
			 	case 'h' :
			 		showResponseHeaders = true;
			 	break;
			 	
			 	case 'k' :
			 		keepLocal = true;
			 	break;
			 	
			 	case 'P' :
			 		showPost = true;
			 	break;
			 	
			 	case 'C' :
			 		reuseCookies = true;
			 		cookieFile = g.getOptarg();
			 	break;
			 	
			 	case 'E' :
			 		strictEncoding = true;
				break;
			 	
			 	case 'S' :
			 		doubleSpace = false;
				break;
				
			 	case 'b' :
			 		brief = true;
				break;
			 	
			 	case ':':
	       		System.err.println("Option " + (char)g.getOptopt() + " requires an argument");
					opterror = true;
				break;
	        
		 		case '?' :
		 			System.err.println("Unrecognized option `" + args[g.getOptind()-1] + "'");
		 			showUsage();
		 			System.exit(1);
		 		break;
		 		
		 		default	:
		 			opterror = true;
				break;
			}
		}
		 
		if(opterror)
		{
System.exit(1);		 	
		}

		try
		{
			summaryLogName = AGENT_NAME + '-' + System.currentTimeMillis() + ".report";
			openSummaryLog(summaryLogName);

			if(proxy != null)
			{
				setProxy(proxy);
			}
			
			if(conf != null)
			{
				DOMParser parser = new DOMParser();
				parser.setFeature(
					"http://xml.org/sax/features/validation",
					validate_config);
				parser.setFeature(
					"http://apache.org/xml/features/nonvalidating/load-dtd-grammar",
					validate_config);
				parser.setFeature(
					"http://apache.org/xml/features/nonvalidating/load-external-dtd",
					validate_config);
						
				InputSource confSource = new InputSource(
						new FileInputStream(conf));
				parser.parse(confSource);
				Document document = parser.getDocument();
				setHeaders(document);
				setAcceptImages(document);
				setContentEncoding(document);
				setIgnores(document);
				setSyntaxFilter(document);
				inputContainer = new InputContainer(document);
			}
			else
			{
				inputContainer = new InputContainer();
			}
			
			if(typeToMime.get(type) == null)
			{
System.err.println("can't handle `" + type + "'");	
				System.exit(1);
			}
			if(g.getOptind() >= args.length)
			{
		 		showUsage();
		 		System.exit(1);
			}
			int w = 0, f = 0, d = 0;
			int nruns = 0;
			for(int i  = g.getOptind(); i < args.length; ++i)
			{
				++nruns;
				WurML worm = new WurML(depth,type,args[i],
					showCookies, dumpContent);
				worm.summaryLog("depth = " + depth + ", type = `" + type + "' ");
				if(agent != null)
				{
					worm.setAgent(agent);
					worm.summaryLog(", User-Agent = `" + agent + "'");
				}
				worm.summaryLog(EOL);

				System.out.println("starting with `" + args[i] + "'");	
				worm.summaryLog("starting with `" + args[i] + "'" + EOL);
				
				worm.traverse(
					new RequestProfile(RequestProfile.GET,new URL(args[i]),args[i]));
				worm.saveCookies();
				d += worm.getAttempts();
				f += worm.getFailures();
				w += worm.getWarnings();
				System.out.println();
				System.out.println(
					args[i] + " traversal summary: " 
					+" urls attempted: " + worm.getAttempts() 
					+ ", failures: " + worm.getFailures()
					+ ", warnings: " + worm.getWarnings()
					);
				System.out.println();
				logWriter.write(EOL
					+ args[i] + " traversal summary: " 
					+" urls attempted: " + worm.getAttempts() 
					+ ", failures: " + worm.getFailures()
					+ ", warnings: " + worm.getWarnings()
					+ EOL + EOL);
			}
			if(nruns > 1)
			{
				System.out.println("");
				System.out.println(
					"total urls attempted: " + d 
					+ ", errors: " + f
					+ ", warnings: " + w
					);
				logWriter.write(EOL
					+ "total urls attempted: " + d 
					+ ", errors: " + f
					+ ", warnings: " + w
					+ EOL);
			}
			logWriter.close();
			System.out.println("summary written to `" + summaryLogName + "'");
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
	}
	static void showUsage()
	{
		System.out.println("Usage: wurml [OPTIONS] url1 [url2 ...]");
		System.out.println("  -c, --config=CONF\tuse specified configuration file");	
		System.out.println("  -d, --depth=INT\ttraverse urls to specified depth (default 3)");	
		System.out.println("  -a, --agent=STRING\tset User-Agent (default 'WurML', overrides config)");
		System.out.println("  -p, --proxy=HOST[:PORT]\tuse HTTP proxy");
		System.out.println("  -s, --show-cookies\tshow cookies sent and received in output");	
		System.out.println("  -P, --show-post\tshow POST arguments");	
		System.out.println("  -h, --show-headers\tshow HTTP response headers");	
		System.out.println("  -m, --dump\t\tdump received documents to output");	
		System.out.println("  -i, --ignore-robots\tignore 'robots.txt' and <META name='robots' ...>");	
		System.out.println("  -r, --revisit\t\twhile traversing, return to previously fetched urls");	
		System.out.println("  -n, --no-validation\tdo not validate against DTD");	
		System.out.println("  -k, --keep-local\tdo not visit links to secondary hosts");	
		System.out.println("  -y, --dry-run\t\tload first document and show urls; do not fetch them");	
		System.out.println("  -l, --lines\t\tprint line numbers in syntax error report");	
		System.out.println("  -x, --trust-config\tdo not validate the configuration file");	
		System.out.println("  -b, --brief\tmake brief report to standard output");	
		System.out.println("  -C, --saved-cookies[=FILE]\tuse stored cookies from named file");
		System.out.println("  -E, --strict-encoding\tuse strict character encoding as per RFC 3023");
		System.out.println("  -S, --single-space\tsuppress new-line betwen urls");
		System.out.println("  -v, --version\t\tshow the version and exit");	
		System.out.println("  -u, --usage\t\tshow this message");	
	}
	
	static class RequestProfile
	{
		public static final String POST="POST";
		public static final String GET="GET";
		
		protected String method;
		protected String url;
		protected List params;
		protected URL parent; 
		
		public RequestProfile(String method, URL parent, String url)
		{
			this.method = method;
			this.url = url;
			this.parent = parent;
			this.params = new ArrayList();
		}
		public void addParam(String n, String v, Node node)
		{
			params.add(new Param(n,v,node));
		}
		public URL createURL()
			throws MalformedURLException
		{
			return newURL(parent,url);
		}
		public int paramHash()
		{
			int hash = 0;
			Iterator it = params.iterator();
			while(it.hasNext())
			{
				hash ^= it.next().hashCode();
			}
			return hash;
		}
		public URL getParent()
		{
			return parent;
		}
		public void setParent(URL parent)
		{
			this.parent = parent;
		}
		public int hashCode()
		{
			int hash = method.hashCode() ^ url.hashCode();
			if(parent != null)
			{
				hash ^= parent.toExternalForm().hashCode(); 
			}
			hash ^= paramHash();
			return hash;
		}
		
		
		public String toExternalForm()
			throws MalformedURLException
		{
			StringBuffer buff = new StringBuffer();
			buff.append(method).append(" - ");
			if(parent == null)
			{
				buff.append(url);
			}
			else
			{
				URL thisurl = newURL(parent,url);
				buff.append(thisurl);
			}
			if(params.size() > 0)
			{
				Param param;
				Iterator it = params.iterator();
				while(it.hasNext())
				{
					param = (Param) it.next();
					buff.append(EOL)
						.append("\t").append(param.name)
						.append('=').append(param.value);
				}
			}
			
			return buff.toString();
		}
		
	}

	static class Param 
	{
		String name;
		String value;
		Node searchNode;
		public Param(String name,String value, Node searchNode)
		{
			this.name = name;
			this.value= value;
			this.searchNode = searchNode;
		}
		public int hashCode()
		{
			return name.hashCode() ^ value.hashCode();
		}
	}
	static class Response
	{
		URL url;
		Document document;
		String encoding;
		ContentType ct = null;
		String text = null;
		Map headers = null;
		public Response(URL url, Document document,String encoding)
		{
			this.url = url;
			this.document = document;
			this.encoding = encoding;
		}
	}

	
}
