/*
 * This software is copyright (c) 2004 by Shadowplay.net. It is released 
 * under the terms of the Artistic License.  Please refer to the file 
 * LICENSE for details.
 */
package net.shadowplay.wurml.input;

import java.io.UnsupportedEncodingException;
import java.net.URL;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import net.shadowplay.wurml.Logger;
import net.shadowplay.wurml.WapConstants;
import net.shadowplay.wurml.WurML;
import net.shadowplay.wurml.filter.InputFilter;
import net.shadowplay.wurml.filter.OptionWithValues;
import net.shadowplay.wurml.filter.SetvarFilter;

import org.apache.xerces.dom.DocumentImpl;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.traversal.NodeFilter;
import org.w3c.dom.traversal.NodeIterator;

/**
 * @author Michael Dykman for Shadowplay.net 
 */


public class SubstString implements Iterable
{
	public static final int MODE_MIXED = 0;
	public static final int MODE_RAW = 1;
	public static final int MODE_VAR = 2;
	

	public static final int VAR_ESC = 0;
	public static final int VAR_UNESC = 1;
	public static final int VAR_NO_ESC = 2;

	protected Map vars = new HashMap();
	protected Map systemVariables;
	private URL url;
	protected List entryList = new ArrayList();
	protected int nVars;
	protected String orig;
	private Logger logger;
	private boolean abort = false;
	private List complaints;
	private String encoding;
	
	private final void init(
		Logger logger, 
		URL url, 
		Map systemVariables,
		List complaints,
		String string,
		String encoding)
	{
		this.logger = logger;
		this.complaints = complaints;
		this.nVars = 0;
		this.orig = string;
		this.systemVariables = systemVariables;
		this.url=url;
		this.encoding = encoding;
	}
	
	public SubstString(Logger logger, URL url, Node node,	Map systemVariables,
		String string, List complaints, String docType, int substType, String encoding)
		throws MalformedVariableException
	{
		init(logger,url,systemVariables,complaints,string, encoding);
		int esc_mode = VAR_NO_ESC;
		
		if(substType == MODE_MIXED)
		{
			int offset = 0;
			int start,end;
			Entry e;
			while(offset < string.length() && (start = string.indexOf('$',offset)) != -1)
			{
				int nextoffset;
				end = start+1;
				if(string.charAt(end) == '$') // '$$' renders to raw '$'
				{
					e = new Entry(string.substring(offset,end),null,VAR_NO_ESC);
					entryList.add(e);
					++end;
					nextoffset = end;
				}
				else 	// variable found
				{
					String id;
					
					int endraw;
					if(string.charAt(end) == '(')
					{
						start = end;
						end = string.indexOf(')',start);
						if(end == -1)
						{
							throw new MalformedVariableException("formal variable found with no terminating ')'");
						}
						endraw = start == 0 ? 0 : start - 1;
						nextoffset = end + 1;
						id = string.substring(start+1,end);
						int off;
						if((off = id.indexOf(':')) != -1)
						{
							if(id.length() > off + 1)
							{
								String suffix = id.substring(off+1);
								id = id.substring(0,off);
								esc_mode = parseEscapeMode(suffix,string);
							}
							else
							{
								throw new MalformedVariableException(
									"variable ends with colon (':') and has no suffix in `"
									+ string + "'");
							}
						}
						else // no suffix
						{
							esc_mode = VAR_NO_ESC;
						}
					}
					else
					{
						if(string.charAt(end) == '_' ||
							Character.isLetter(string.charAt(end)))
						{
							while(end < string.length() && 
								Character.isLetterOrDigit(string.charAt(end)))
								++end;
							endraw = start;
							nextoffset = end;
							id = string.substring(start+1,end);
							esc_mode = VAR_NO_ESC;
						}
						else
						{
							throw new MalformedVariableException(
								"variable must start with a letter or underscore ('_') in `"
								+ string + "'");
						}
					}
					++nVars;
					findValues(node,id,docType);
					String raw = "";
					if(endraw > offset)
					{
						raw = string.substring(offset,endraw);
					}
					e = new Entry(raw,id,esc_mode);
					addEntry(e);
					offset = end;
				}
				offset = nextoffset;
			}
			if(! (offset >= string.length()))
			{
				addEntry(new Entry(string.substring(offset),null,VAR_NO_ESC));
			}
		}
		else if(substType == MODE_RAW)
		{
			addEntry(new Entry(string,null,VAR_NO_ESC));
			findValues(node,string,docType);
		}
		else // VAR
		{
			addEntry(new Entry("",string,VAR_NO_ESC));
		}
	}
	
	private static int parseEscapeMode(String suffix, String string)
		throws MalformedVariableException
	{
		int mode; 
		if(suffix.equals("e") || suffix.equals("escape"))
		{
			mode = VAR_ESC;
		}
		else if(suffix.equals("u") || suffix.equals("unesc"))
		{
			mode = VAR_UNESC;
		}
		else if(suffix.equals("n") || suffix.equals("noesc"))
		{
			mode = VAR_NO_ESC;
		}
		else
		{
			throw new MalformedVariableException("unrecognized variable suffix `" 
				+ suffix + "' found in `" + string + "'");
		}
		return mode;
	}
	private void addEntry(Entry e)
	{
		entryList.add(e); 
	}
	public String getSubst(String [] subs)
	{
		StringBuffer b = new StringBuffer();
		Iterator it = entryList.iterator();
		int i = 0;
		while(it.hasNext())
		{
			Entry e = (Entry)it.next();
			b.append(e.literal);
			if(e.var != null)
			{
				b.append(subs[i++]);
			}
		}
		
		return b.toString();
	}

	private void findValues(Node doc, String id,String docType)
	{
		List values;
		if(docType.equals(WapConstants.WML_TYPE))
		{
			values = findValuesForWML(doc,id);
		}
		else // if(docType.equals(WapConstants.WML_TYPE))
		{
			values = findValuesForXHTML(doc,id);
		}
		vars.put(id,values);
	}
	private List findValuesForXHTML(Node doc, String id)
	{
		List values = new ArrayList();
		DocumentImpl impl = new DocumentImpl();
		NodeIterator it = impl.createNodeIterator(
			doc,NodeFilter.FILTER_ACCEPT,new OptionWithValues(id));
		Node n;
		NamedNodeMap attr;
		
		while((n = it.nextNode()) != null)
		{
			attr = n.getAttributes();
			values.add(attr.getNamedItem("value").getNodeValue());
		}
		return values;
	}
	private List findValuesForWML(Node doc, String id)
	{
		List values = new ArrayList();
		Node n;
		NamedNodeMap attr;
		
		int inputs = 0;
		DocumentImpl impl = new DocumentImpl();
		NodeIterator it = impl.createNodeIterator(
			doc,NodeFilter.FILTER_ACCEPT,new OptionWithValues(id));
		while((n = it.nextNode()) != null)
		{
			attr = n.getAttributes();
			values.add(attr.getNamedItem("value").getNodeValue());
			inputs = 1;
		}
		
		if(values.size() == 0)
		{
			it = impl.createNodeIterator(
				doc,NodeFilter.FILTER_ACCEPT,new InputFilter(id));
			List predefs = null;
			while((n = it.nextNode()) != null)
			{
				attr = n.getAttributes();
				if(attr.getNamedItem("emptyok") != null)
				{
					values.add("");
				}

				predefs = WurML.inputContainer.getValues(url.toExternalForm(),id);
				values.addAll(predefs);
				
				if(inputs == 0 && values.size() == 0)
				{
					abort = true;
					if(!complaints.contains(id))
					{
						complaints.add(id);
						StringBuffer buff = new StringBuffer();
						buff.append("in document ").append(url.toExternalForm())
							.append(": no configured values to use for ")
							.append("<INPUT name = '").append(id).append("' ");
						
						String maxlen = null;
						String format = null;
						Node att;
						attr = n.getAttributes();
						if((att = attr.getNamedItem("maxlength")) != null)
						{
							maxlen = att.getNodeValue();
						}
						if((att = attr.getNamedItem("format")) != null)
						{
							format = att.getNodeValue();
						}
						if(format != null)
						{
							buff.append("format='" + format + "' ");
						}
						if(maxlen != null)
						{
							buff.append("maxlength='" + maxlen + "' ");
						}
						buff.append('>');
						
						logger.warning(buff.toString());
					}
				}
				++inputs;
			}
			if(inputs > 1)
			{
				if(!complaints.contains(id))
				{
					complaints.add(id);
					logger.warning(inputs + " Input nodes found with id = `" + id + "'");
				}
			}
		}
		
		if(values.size() == 0)
		{
			it = impl.createNodeIterator(
				doc,NodeFilter.FILTER_ACCEPT,new SetvarFilter(id));
			while((n = it.nextNode()) != null)
			{
				attr = n.getAttributes();
				values.add(attr.getNamedItem("value").getNodeValue());
			}
		}
		
		if(abort == false && inputs == 0 && values.size() == 0)
		{
			Object o = systemVariables.get(id);
			if(o == null)
			{
				if(!complaints.contains(id))
				{
					complaints.add(id);

					logger.warning("no variable found with id `" + id 
						+ "' in document `" + url.toExternalForm() + "'");
				}
// abort this one				
				abort = true;
			}
			else
			{
				values.add(o);
			}
		}
		return values;
	}
	public String escape(String in)
		throws UnsupportedEncodingException
	{
		return URLEncoder.encode(in,encoding);
	}
	public String unescape(String in)
		throws UnsupportedEncodingException
	{
		return URLDecoder.decode(in,encoding);
	}
	public Iterator iterator()
//		throws UnsupportedEncodingException
	{
		Iterator iterator;
		if(abort)
		{
			iterator = new SingletonIterator();
		}
		else if(nVars == 0)
		{
			iterator = new SingletonIterator(orig);
		}
		else
		{
			iterator = new SubstIterator();
		}
		return iterator;
	}
	
	static class Entry
	{
		String literal;
		String var;
		int esc_mode;
		public Entry(String literal,String var, int esc_mode)
		{
			this.literal = literal;
			this.var = var;
			this.esc_mode = esc_mode;
		}
	}
	
	protected class SingletonIterator implements Iterator
	{
		boolean used = false;
		Object o;
		public SingletonIterator()
		{
			used = true;
		}
		public SingletonIterator(Object o)
		{
			this.o = o;
		}
		public boolean hasNext()
		{
			if(used == false)
			{
				return true;
			}
			else
			{
				return false;
			}
		}
		public Object next()
		{
			used = true;
			return o;
		}
		public void remove()
		{
			throw new UnsupportedOperationException("this iterator does not support the remove() method");
		}

	}
	protected class SubstIterator implements Iterator
	{
		int[] indices;
		List[] lists;
		String [] names;
		public SubstIterator()
//			throws UnsupportedEncodingException
		{
			if(nVars > 0)
			{
				indices = new int[nVars];
				lists = new List[nVars];
				names = new String[nVars];
				
				for(int i = 0; i < nVars; ++i)
				{
					indices[i] = 0;
				}
				Iterator it = entryList.iterator();
				int i = 0;
				while(it.hasNext())
				{
					Entry e = (Entry) it.next();
					if(e.var != null)
					{
						List vv,pvv = new ArrayList();
						names[i] = e.var;
						vv = (List)vars.get(e.var);
// do escape or unescapes on all values if appropriate
						if(e.esc_mode != VAR_NO_ESC)
						{
							Iterator vit = vv.iterator();
							try
							{
								while(vit.hasNext())
								{
									if(e.esc_mode == VAR_ESC)
									{
										pvv.add(escape((String) vit.next()));
									}
									else // e.esc_mode == VAR_UNESC
									{
										pvv.add(unescape((String) vit.next()));
									}
								}
							}
							catch(UnsupportedEncodingException ee)
							{
								throw new RuntimeException(ee);
							}
							lists[i] = pvv;
						}
						else
						{
							lists[i] = vv;
						}
						++i;
					}
				}
			}
			else
			{
				indices = null;
			}
		}
		
		public boolean hasNext()
		{
			boolean result = false;
			if(indices != null)
			{
				result = true;
			}
			return result;
		}
		public Object next()
		{
			Object result;
			if(nVars == 0)
			{
				result = orig;
				indices = null;
			}
			else
			{
				String [] subs = new String[nVars];

				for(int i = 0; i < nVars; ++i)
				{
					List l = lists[i];
					subs[i] = (String)l.get(indices[i]);
					systemVariables.put(names[i],subs[i]);
				}
				++indices[nVars-1];
				for(int i = nVars-1; i >= 0; --i)
				{
					if(indices[i]>= lists[i].size())
					{
						if(i > 0)
						{
							indices[i] = 0;
							indices[i-1]++;
						}
						else
						{
							indices = null;
						}
					}
				}
				result = getSubst(subs);
			}
			return result;
		}
		
		public void remove()
		{
			throw new UnsupportedOperationException("this iterator does not support the remove() method");
		}
	}
}
